theory Setconstr imports Main Boolean_algebra

  (* Like Setconstr.thy, but variables are not embedded in atoms, 
  but are a separate constructor. *)

begin


notation
  less_eq  (infix "\<sqsubseteq>" 50) and
  less (infix "\<sqsubset>" 50) and
  inf  (infixl "\<sqinter>" 70) and
  sup  (infixl "\<squnion>" 65) and
  top ("\<top>") and
  bot ("\<bottom>")


datatype bop = Sup |Inf
datatype bconst = Top | Bot

lemma not_Inf_Sup [simp]: "(b \<noteq> Inf) = (b = Sup)"
by (case_tac b) simp+
lemma not_Sup_Inf [simp]: "(b \<noteq> Sup) = (b = Inf)"
by (case_tac b) simp+

fun
  neutral_bop ::"bop \<Rightarrow> bconst"
where
  "neutral_bop Sup = Bot"
| "neutral_bop Inf = Top"


  (* representation of sets *)
datatype ('n, 'a) set_rep = 
    Atom 'a                   	-- "Atom"
  | Bconst bconst               -- "Constant"
  | Var 'n                      -- "Variable"
  | Comp "('n, 'a) set_rep"     -- "Complement"
  | Binop bop "(('n, 'a) set_rep)" "(('n, 'a) set_rep)"   -- "Binary operation"


  (* substitution in set_rep *)

fun
  subst_set_rep :: "('n \<Rightarrow> ('n, 'a) set_rep option) \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "subst_set_rep s (Atom a) = (Atom a)"
| "subst_set_rep s (Bconst b) = (Bconst b)"
| "subst_set_rep s (Var n) = (case (s n) of None \<Rightarrow> (Var n) | Some t \<Rightarrow> t)"
| "subst_set_rep s (Comp e) = (Comp (subst_set_rep s e))"
| "subst_set_rep s (Binop bop e e') = (Binop bop (subst_set_rep s e) (subst_set_rep s e'))"


  (* interpretation of set_rep *)

fun
  interp_bop  :: "bop \<Rightarrow> 'a::boolean_algebra\<Rightarrow>'a\<Rightarrow>'a"
where 
  "interp_bop Sup = sup"
| "interp_bop Inf = inf"

fun
  interp_bconst :: " bconst \<Rightarrow> 'a::complete_boolean_algebra"
where
  "interp_bconst Top = top"
| "interp_bconst Bot = bot"


fun
  interp_set_rep :: "('a \<Rightarrow> 'b::complete_boolean_algebra) \<Rightarrow> ('n \<Rightarrow> 'b)\<Rightarrow> ('n, 'a) set_rep \<Rightarrow> 'b"
where 
  "interp_set_rep ai vi (Atom a) = ai a"
| "interp_set_rep ai vi (Bconst c) = interp_bconst c"
| "interp_set_rep ai vi (Var n) = vi n"
| "interp_set_rep ai vi (Comp a) = - (interp_set_rep ai vi a)"
| "interp_set_rep ai vi (Binop bop a1 a2) = 
       (interp_bop bop) (interp_set_rep ai vi a1)(interp_set_rep ai vi a2)"


subsection{*Transformation in DNF and CNF *}

subsubsection{*Distributing the complement*}

text
{*  Example : @{text "comp_distrib Comp (Sup a b)=(Inf (Comp a) (Comp b))"}
*}

fun
  invert_bconst :: "bconst \<Rightarrow> bconst"
where
  "invert_bconst Top = Bot"
| "invert_bconst Bot = Top"

fun
  invert_bop ::"bop \<Rightarrow> bop"
where
  "invert_bop Sup = Inf"
| "invert_bop Inf = Sup"

fun
  comp_distrib :: "('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where 
    "comp_distrib (Atom a) = (Atom a)"
  | "comp_distrib (Bconst b) = (Bconst b)"
  | "comp_distrib (Var n) = (Var n)"

  | "comp_distrib (Comp (Atom a)) = (Comp (Atom a))"
  | "comp_distrib (Comp (Bconst b)) = (Bconst (invert_bconst b))"
  | "comp_distrib (Comp (Var n)) = (Comp (Var n))"

  | "comp_distrib (Comp (Comp a))=(comp_distrib a)"
  | "comp_distrib (Comp  (Binop bop  a b )) =
     (Binop (invert_bop bop) (comp_distrib(Comp a))(comp_distrib(Comp b)))"

  | "comp_distrib (Binop bop a b )= (Binop bop (comp_distrib a) (comp_distrib b))"



fun
  is_elementary :: "('n, 'a) set_rep \<Rightarrow> bool"
where
  "is_elementary e = 
    (case e of
      (Atom a) => True
    | (Bconst b) => True
    | (Var n) => True
    | _ => False)"


fun 
  is_comp_distrib :: "('n, 'a) set_rep \<Rightarrow> bool"
where
  "is_comp_distrib (Atom a) = True"
| "is_comp_distrib (Bconst b) = True"
| "is_comp_distrib (Var n) = True"
| "is_comp_distrib (Comp a)= (is_elementary a)"
| "is_comp_distrib (Binop bop a b) = (is_comp_distrib a \<and> is_comp_distrib b)"


lemma comp_invert_bconst [simp]:
  "interp_bconst (invert_bconst b)= -(interp_bconst b)"
by (case_tac b) simp+


lemma comp_distrib_is_comp_distrib: "is_comp_distrib (comp_distrib f)"
by (induct_tac f rule: comp_distrib.induct) simp_all


lemma comp_interp_bop [simp]:
  "interp_bop (invert_bop x) (- a) (- b) = -(interp_bop x a b)"
by (case_tac x) simp+


lemma interp_comp_distrib [simp]:
  "interp_set_rep ai vi (comp_distrib a) = interp_set_rep ai vi a"
by (induct a rule: comp_distrib.induct) simp_all

fun 
  is_elementary_composition :: "bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> bool"
where
  "is_elementary_composition sop (Atom a) = True"
| "is_elementary_composition sop (Bconst b) = True"
| "is_elementary_composition sop (Var n) = True"
| "is_elementary_composition sop (Comp a) = (is_elementary a)"
| "is_elementary_composition sop (Binop bop a b) = 
    ((bop = sop) \<and> is_elementary_composition sop a \<and> is_elementary_composition sop b)"

fun
  is_distrib :: "bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> bool"
where
  "is_distrib sop (Atom a) = True"
| "is_distrib sop (Bconst b) = True"
| "is_distrib sop (Var n) = True"
| "is_distrib sop (Comp a)= (is_elementary a)"
| "is_distrib sop (Binop bop a b)= 
  (if bop=sop 
  then
  (is_elementary_composition sop a) \<and> (is_elementary_composition sop b)
  else
  (is_distrib sop a) \<and> (is_distrib sop b))"

lemma elementary_composition_impl_distrib [rule_format]:
  "is_elementary_composition bop a \<longrightarrow> is_distrib bop a"
by (induct a) auto


lemma distrib_impl_comp_distrib [rule_format]:
  "is_distrib bop a \<longrightarrow> is_comp_distrib a"
by (induct a) (auto simp add: elementary_composition_impl_distrib)+


fun 
  distrib1 :: " bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "distrib1 sop f (Atom a) = (Binop sop  f (Atom a))"
| "distrib1 sop f (Bconst b) = (Binop sop f (Bconst b))"
| "distrib1 sop f (Var n) = (Binop sop  f (Var n))"
| "distrib1 sop f (Comp a) = (Binop sop  f (Comp a))"
| "distrib1 sop f (Binop bop a b)=
  (if bop=sop
  then
  (Binop sop f (Binop bop a b))
  else
  (Binop bop (distrib1 sop f a) (distrib1 sop f b)))"


lemma interp_bop_distrib: "\<lbrakk>bop\<noteq>bopa\<rbrakk>\<Longrightarrow> 
  (interp_bop bop) 
    (interp_bop bopa (interp_set_rep ai vi f) (interp_set_rep ai vi g1))
    (interp_bop bopa (interp_set_rep ai vi f) (interp_set_rep ai vi g2)) =
  (interp_bop bopa) (interp_set_rep ai vi f)
        (interp_bop bop (interp_set_rep ai vi g1)(interp_set_rep ai vi g2))"
apply (case_tac bop)
apply (case_tac bopa)
apply (simp add: sup_inf_distrib1)
apply (simp add: inf_sup_distrib1)
apply (case_tac bopa)
apply (simp add: sup_inf_distrib1)
apply (simp)
done


lemma interp_distrib1: 
  "interp_set_rep ai vi (distrib1 bop f g) 
  = (interp_bop bop) (interp_set_rep ai vi f)  (interp_set_rep  ai vi g)"
apply (induct g)
apply (simp add: inf_sup_distrib2 inf_sup_distrib1 interp_bop_distrib)+
done


lemma distrib1: "\<lbrakk>is_elementary_composition sop f; is_distrib sop g\<rbrakk> \<Longrightarrow>
  is_distrib sop (distrib1 sop f g)"
by (induct g) clarsimp+


lemma comp_distrib1 : "\<lbrakk>is_comp_distrib f ; is_comp_distrib g\<rbrakk> \<Longrightarrow>
  is_comp_distrib (distrib1 sop f g)"
by (induct g) clarsimp+

fun 
  distrib2 :: "bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "distrib2 sop (Atom a) b =(distrib1 sop (Atom a) b)"
| "distrib2 sop (Comp a) b =(distrib1 sop (Comp a) b)"
| "distrib2 sop (Var n) b =(distrib1 sop (Var n) b)"
| "distrib2 sop (Bconst a) b = distrib1 sop (Bconst a) b"
| "distrib2 sop (Binop bop a1 a2) b = 
  (if bop=sop 
  then (distrib1 sop (Binop bop a1 a2) b)
  else (Binop bop (distrib2 sop a1 b) (distrib2 sop a2 b)))"

lemma comp_distrib2 : "\<lbrakk>is_comp_distrib f; is_comp_distrib g \<rbrakk> \<Longrightarrow>
  is_comp_distrib (distrib2 sop f g)"
by (induct f) (clarsimp simp add: comp_distrib1)+

lemma interp_distrib2:
  "interp_set_rep ai vi (distrib2 sop f g) = 
  interp_bop sop (interp_set_rep ai vi f) (interp_set_rep ai vi g)"
apply (induct f)
apply (simp add: interp_distrib1 inf_sup_distrib2 interp_bop_distrib)+
  (* case Binop *)
apply (rename_tac bop f1 f2)
apply (case_tac bop)
apply (case_tac sop)
apply simp
apply (simp add: inf_sup_distrib2)
apply (case_tac sop)
apply (simp add: sup_inf_distrib2)
apply simp
done

lemma  distrib2 :"\<lbrakk>is_distrib sop f; is_distrib sop g\<rbrakk>\<Longrightarrow>
  is_distrib sop (distrib2 sop f g)"
by (induct f) (clarsimp simp add: distrib1)+

fun 
  distrib :: "bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "distrib sop (Atom a) = (Atom a)"
| "distrib sop (Bconst b) = (Bconst b)"
| "distrib sop (Var n) = (Var n)"
| "distrib sop (Comp a) = (Comp a)"
| "distrib sop (Binop bop a b) =
  (if bop=sop
  then 
  (distrib2 sop (distrib sop a) (distrib sop b))
  else 
  (Binop bop (distrib sop a) (distrib sop b)))"


lemma interp_set_rep_distrib [simp]: 
  "interp_set_rep ai vi (distrib sop f) = interp_set_rep ai vi f"
by (induct f) (clarsimp simp add: interp_distrib2)+


lemma distrib_is_distrib : "is_comp_distrib f \<Longrightarrow> is_distrib sop (distrib sop f)"
by (induct f) (clarsimp simp add: distrib2)+


lemma distrib_is_comp_distrib :
"is_comp_distrib f \<Longrightarrow> is_comp_distrib (distrib sop f)"
by (induct f) (clarsimp simp add: comp_distrib2)+

  (* ------------------------------------------------------------ *)
  (* DNF and CNF  *)
  (* ------------------------------------------------------------ *)



definition
  dnf :: "('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep" where
  "dnf a = (distrib Inf (comp_distrib a))"
definition
  is_dnf :: "('n, 'a) set_rep \<Rightarrow> bool" where
  "is_dnf a = is_distrib Inf a"
definition
  cnf ::"('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep" where
  "cnf f = (distrib Sup (comp_distrib f))"
definition
  is_cnf :: "('n, 'a) set_rep \<Rightarrow> bool" where
  "is_cnf f = is_distrib Sup f"

lemma dnf_is_dnf [simp]: "is_dnf (dnf a)"
by (simp add: is_dnf_def dnf_def)
   (simp add: comp_distrib_is_comp_distrib distrib_is_distrib distrib_is_comp_distrib)

lemma cnf_is_cnf [simp]: "is_cnf (cnf f)"
by (simp add: is_cnf_def cnf_def)
   (simp add: comp_distrib_is_comp_distrib distrib_is_distrib distrib_is_comp_distrib)


lemma interp_set_rep_dnf [simp]:
  "interp_set_rep ai vi (dnf f) = interp_set_rep ai vi f"
by (simp add: dnf_def)


lemma interp_set_rep_cnf [simp]:
  "interp_set_rep ai vi (cnf f)=interp_set_rep ai vi f"
by (simp add: cnf_def)

   (******************************************************************)
   (*              Simplifications                                   *)
   (******************************************************************)

fun 
  is_bconst :: "('n, 'a) set_rep \<Rightarrow> bool"
where
  "is_bconst (Atom a) = False"
| "is_bconst (Var n )= False"
| "is_bconst (Bconst b)=True"
| "is_bconst (Comp a)=False"
| "is_bconst (Binop sop a b)=False"



fun
  simplif_set_rep1  :: "('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "simplif_set_rep1  (Binop sop a b)=
    (case sop of
      Sup \<Rightarrow> 
       (if (a = (Bconst Top) \<or>  b = (Bconst Top))
        then (Bconst Top)
        else (Binop sop a b))
     | Inf \<Rightarrow> 
       (if (a = (Bconst Bot) \<or> b =( Bconst Bot))
        then (Bconst Bot)
        else (Binop sop a b)))"
| "simplif_set_rep1 x = x"
  
lemma simplif_set_rep1_equiv [rule_format,simp]:
  "interp_set_rep ai vi (simplif_set_rep1 a)= interp_set_rep ai vi a"
by (case_tac a) (simp split add:  bop.splits)+

fun
  simplif_set_rep2  :: "('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep"
where
  "simplif_set_rep2  (Binop sop a b) =
   (case sop of 
     Sup \<Rightarrow> 
      (if (a = Bconst Bot)
       then b
       else if (b = Bconst Bot)
       then a
       else (Binop sop a b))
    | Inf \<Rightarrow> 
     (if (a = Bconst Top)
      then b
      else if (b = Bconst Top)
      then a
      else (Binop sop a b)))"
 | "simplif_set_rep2 x = x"


lemma simplif_set_rep2_equiv [rule_format,simp]: 
  "interp_set_rep ai vi (simplif_set_rep2 a)=interp_set_rep ai vi a"
by (case_tac a) (simp split add:  bop.splits)+

fun 
  simplif_set_rep :: "('n,'a) set_rep \<Rightarrow> ('n,'a) set_rep"
where
  "simplif_set_rep (Atom a) = (Atom a)"
| "simplif_set_rep (Bconst b)=(Bconst b)"
| "simplif_set_rep (Var n) = (Var n)"
| "simplif_set_rep (Comp a) = (Comp (simplif_set_rep  a))"
|  "simplif_set_rep (Binop sop a b)=
  (simplif_set_rep2 (simplif_set_rep1 (Binop sop (simplif_set_rep  a)(simplif_set_rep   b))))"



lemma simplif_set_rep_equiv :
  "interp_set_rep  ai vi (simplif_set_rep  a) = interp_set_rep ai vi a"
by (induct a)
   (fastforce split add: bop.split)+


fun 
  no_bconst_in_binop  :: "('n, 'a) set_rep \<Rightarrow> bool"
where
  "no_bconst_in_binop  (Atom a)=True"
| "no_bconst_in_binop  (Bconst b)=True"
| "no_bconst_in_binop  (Var n)=True"
| "no_bconst_in_binop  (Comp a)=(no_bconst_in_binop a)"
| "no_bconst_in_binop  (Binop sop a b)=
  (if ((is_bconst a) \<or> (is_bconst b))
  then
    False
  else 
    (no_bconst_in_binop a) \<and> (no_bconst_in_binop b))"





lemma bconst_idem:
  "no_bconst_in_binop (simplif_set_rep (Binop bop (Bconst c) (Bconst c)))"
apply (case_tac c)
apply (case_tac bop)
apply (simp_all)
apply (simp split: bop.split)
done


lemma is_bconst_simplif:
  "\<lbrakk>simplif_set_rep a \<noteq> Bconst Bot;simplif_set_rep a \<noteq> Bconst Top\<rbrakk>\<Longrightarrow> \<not> (is_bconst (simplif_set_rep a))"
apply (case_tac "simplif_set_rep a")
apply (simp_all)
apply (case_tac bconst)
apply (simp)
apply (simp)
done

lemma no_bconst_in_binop_set_rep:
  "no_bconst_in_binop (simplif_set_rep a)"
apply (induct a)
apply (fastforce)+
apply (simp)
apply (intro conjI impI)
apply (simp add: bconst_idem)
apply (fastforce split add: bop.split)+
apply (split  bop.split)
apply (simp_all add: is_bconst_simplif)
done


lemma comp_distrib_elementary:
  "is_comp_distrib (Comp a)\<Longrightarrow>is_elementary a"
apply (induct a)
apply (fastforce)+
done


lemma elementary_simp_idem:
  "is_elementary f\<Longrightarrow> simplif_set_rep f=f"
apply (induct f)
apply (fastforce)+
done
lemma is_elementary_simplif:
  "is_elementary f\<Longrightarrow>is_elementary (simplif_set_rep f)"
apply (induct f)
apply (simp_all)
done

lemma elementary_compos_impl_elementary_compo_simplif:
  "is_elementary_composition bop f\<Longrightarrow>is_elementary_composition bop (simplif_set_rep f)"
apply (induct f)
apply (fastforce)+
apply (simp only :simplif_set_rep.simps is_elementary_composition.simps is_elementary_simplif)
apply (simp)
apply (intro conjI impI)
apply (simp_all split: bop.split)
apply (fastforce)+
done



lemma is_comp_distrib_simplif:
"is_comp_distrib f\<Longrightarrow>is_comp_distrib (simplif_set_rep f)"
apply (induct f)
apply (fastforce)+
apply (subgoal_tac "is_elementary f")
prefer 2 apply (simp add: comp_distrib_elementary)
apply (simp add:elementary_simp_idem)
apply (simp split:  bop.split)
done





lemma is_distrib_simplif:
  "is_distrib bop f\<Longrightarrow> is_distrib bop (simplif_set_rep f)"
apply (induct f)
apply (fastforce)+
apply (case_tac bop)
apply (simp_all only: is_distrib.simps simplif_set_rep.simps is_elementary_simplif)

apply (case_tac "bopa=bop")
apply (simp_all  add: simplif_set_rep.simps)
apply (intro conjI impI)
apply (clarsimp)
apply (simp split: bop.split)
apply (clarsimp)

apply (simp_all split: bop.split)
apply (clarsimp)
apply (simp only: elementary_composition_impl_distrib)
apply (clarsimp)
apply (simp only: elementary_composition_impl_distrib)
apply (clarsimp)
apply (simp only: elementary_composition_impl_distrib)
apply (clarsimp)
apply (simp only: elementary_composition_impl_distrib)
apply (intro conjI  impI)
apply (clarsimp)
apply (simp_all add: elementary_compos_impl_elementary_compo_simplif)
apply (intro conjI impI)
apply (simp_all add: elementary_compos_impl_elementary_compo_simplif)
done





lemma is_dnf_simplif:
  "is_dnf f\<Longrightarrow> is_dnf (simplif_set_rep f)"
by (simp add: is_dnf_def is_comp_distrib_simplif is_distrib_simplif)


lemma is_cnf_simplif:
  "is_cnf f\<Longrightarrow> is_cnf (simplif_set_rep f)"
by (simp add: is_cnf_def is_comp_distrib_simplif is_distrib_simplif)


  (* ------------------------------------------------------------ *)
  (* Set constraints  *)
  (* ------------------------------------------------------------ *)

datatype ('n, 'a) setconstr =
  Lesseq "(('n, 'a) set_rep)" "(('n, 'a) set_rep)"

fun
(*
  interp_constr ::"('a \<Rightarrow> 'b::{complete_boolean_algebra, nontriv}) \<Rightarrow> ('n \<Rightarrow> 'b) \<Rightarrow> ('n, 'a) setconstr \<Rightarrow> bool"
*)
  interp_constr ::"('a \<Rightarrow> 'b::complete_boolean_algebra) \<Rightarrow> ('n \<Rightarrow> 'b) \<Rightarrow> ('n, 'a) setconstr \<Rightarrow> bool"
where
  "interp_constr ai vi (Lesseq e1 e2) =
    (interp_set_rep ai vi e1 \<sqsubseteq> interp_set_rep ai vi e2)"


definition
  true_constr :: "('n, 'a) setconstr" where
  "true_constr =(Lesseq (Bconst Bot) (Bconst Top))"

lemma interp_true_constr [simp]: "interp_constr ai vi true_constr"
by (simp add: true_constr_def)



definition
  valid_constr :: "('a \<Rightarrow> 'b::complete_boolean_algebra) \<Rightarrow> ('n, 'a) setconstr \<Rightarrow> bool" where
  "valid_constr ai c \<equiv> \<forall> vi. interp_constr ai vi c"

lemma valid_true_constr [simp]: "valid_constr ai true_constr"
by (simp add: valid_constr_def)


fun
  nf_constr :: "('n, 'a) setconstr \<Rightarrow>('n, 'a) setconstr"
where "nf_constr (Lesseq f g) = (Lesseq (dnf f) (cnf g))"

fun is_nf ::"('n, 'a) setconstr \<Rightarrow> bool"
where "is_nf (Lesseq f g)= ((is_dnf f)\<and>(is_cnf g))"

fun is_elementary_constr ::"('n, 'a) setconstr \<Rightarrow> bool"
where 
  "is_elementary_constr (Lesseq f g) = 
  (is_elementary_composition Inf f \<and> is_elementary_composition Sup g)"


lemma is_nf_nf_constr [simp]: "is_nf (nf_constr c)"
by (case_tac c, simp)


lemma interp_constr_nf_constr [simp]:
  "interp_constr ai vi (nf_constr c) = interp_constr ai vi c"
by (case_tac c, simp)


fun
  decompose :: "bop \<Rightarrow> ('n, 'a) set_rep \<Rightarrow> ('n, 'a) set_rep list"
where
  "decompose sop (Atom x) = [(Atom x)]"
| "decompose sop (Bconst b) = [(Bconst b)]"
| "decompose sop (Var n) = [(Var n)]"
| "decompose sop (Comp a) = [(Comp a)]"
| "decompose sop (Binop bop  e1 e2) = 
     (if bop=sop
      then (decompose sop e1)@(decompose sop e2)
      else [(Binop bop e1 e2)])"


lemma decompose_is_elementary_composition [rule_format]:
  "\<forall> c \<in> set (decompose (invert_bop bop) sr). 
  is_distrib bop sr \<longrightarrow> 
  is_elementary_composition bop c"
apply (induct sr)
apply simp+
apply (case_tac bop)
apply (clarsimp simp add: ball_Un)+
done


fun 
  setconstr_decompose_left :: "('n, 'a) setconstr \<Rightarrow> ('n, 'a) setconstr list" 
where
  "setconstr_decompose_left (Lesseq f g) = (map (\<lambda> x . (Lesseq x g)) (decompose Sup f))"


lemma dnf_setconstr_decompose_left [rule_format]:
  "(\<forall> c \<in> set (setconstr_decompose_left sc) .interp_constr ai vi c)= interp_constr ai vi sc"
apply (case_tac sc)
apply (clarsimp)
apply (rename_tac f g)
apply (induct_tac f)
apply (simp add: is_dnf_def ivl_disj_un ball_Un)+
done

lemma setconstr_decompose_left_is_elementary_composition [rule_format]:
  "is_nf sc \<Longrightarrow> 
  \<forall> c \<in> set (setconstr_decompose_left sc). 
  (\<forall> f g. (c = (Lesseq f g) \<longrightarrow> is_elementary_composition Inf f))"
by (case_tac sc) (clarsimp simp add: decompose_is_elementary_composition is_dnf_def)


fun 
  setconstr_decompose_right :: "('n, 'a) setconstr  \<Rightarrow>('n, 'a) setconstr list"
where 
  "setconstr_decompose_right (Lesseq f g) = map (\<lambda> x . Lesseq f x) (decompose Inf g)"

lemma cnf_setconstr_decompose_right [rule_format]:"
  (\<forall> c \<in> set (setconstr_decompose_right l) .interp_constr ai vi c)= (interp_constr ai vi l)"
apply (case_tac l)
apply (rename_tac f g)
apply (clarsimp)
apply (induct_tac g)
apply (simp add: is_cnf_def ivl_disj_un ball_Un)+
done


definition 
  setconstr_to_list ::"('n, 'a) setconstr \<Rightarrow> ('n, 'a) setconstr list" where
  "setconstr_to_list c = concat (map (\<lambda> x . setconstr_decompose_right x) (setconstr_decompose_left c))"

lemma interp_constr_setconstr_to_list[simp]:
  "(\<forall> c \<in> set (setconstr_to_list l) . interp_constr ai vi c) = (interp_constr ai vi l)"
apply (simp add: setconstr_to_list_def del: setconstr_decompose_left.simps)
apply (simp add: dnf_setconstr_decompose_left cnf_setconstr_decompose_right)
done

definition 
  decomp_setconstr_list :: "('n, 'a) setconstr list\<Rightarrow>('n, 'a) setconstr list" where
  "decomp_setconstr_list clist =
  concat (map (\<lambda> x . setconstr_to_list (nf_constr x)) clist)"



lemma setconstr_decompose_right_is_elementary_composition:
  "\<lbrakk> (Lesseq l r) \<in> set (setconstr_decompose_right (Lesseq lc rc));
  is_distrib Sup rc \<rbrakk>
  \<Longrightarrow> is_elementary_composition Sup r"
by (clarsimp simp add: decompose_is_elementary_composition)
  (* 
lemma setconstr_decompose_right_is_elementary_composition [rule_format]:
  "is_nf sc \<Longrightarrow> 
  \<forall> c \<in> set (setconstr_decompose_right sc). 
  (\<forall> f g. (c = (Lesseq f g) \<longrightarrow> is_elementary_composition Sup g))"
by (case_tac sc) (clarsimp simp add: decompose_is_elementary_composition is_cnf_def)
 *)

lemma setconstr_to_list_is_elementary_constr [rule_format]: 
  "\<forall> c \<in> set (setconstr_to_list sc). is_nf sc \<longrightarrow> is_elementary_constr c"
apply (case_tac sc)
apply (simp add: setconstr_to_list_def del: setconstr_decompose_left.simps)
apply (clarsimp simp add: image_iff)
apply (simp add: is_dnf_def is_cnf_def decompose_is_elementary_composition)
done


lemma decomp_setconstr_list_is_elementary_constr [rule_format]:
  "\<forall> c \<in> set (decomp_setconstr_list clist). is_elementary_constr c"
apply (simp add: decomp_setconstr_list_def)
apply (induct clist)
apply clarsimp
apply (simp add: setconstr_to_list_is_elementary_constr)
done

(*****************************************************************************)
(*              constraints transformations                                  *)
(*****************************************************************************)

fun 
  constr_switch_left :: "('n,'a) setconstr \<Rightarrow> ('n,'a) setconstr"
where
  "constr_switch_left (Lesseq l r) = (Lesseq (Binop Inf l (Comp r)) (Bconst Bot))"



lemma interp_constr_switch_left [simp]:
  " interp_constr ai vi (constr_switch_left c)=interp_constr ai vi c"
apply (case_tac c)
apply (simp add: inf_ge_compl_bot_iff [THEN sym])
done

lemma valid_constr_switch_left:
  "valid_constr ai c = valid_constr ai (constr_switch_left c)"
by (simp add: valid_constr_def)

fun
  no_right_constr :: "('n,'a) setconstr \<Rightarrow> bool"
where
  "no_right_constr (Lesseq f g) =(g=(Bconst Bot))"


definition
switch_and_nf :: "('n,'a) setconstr \<Rightarrow> ('n,'a) setconstr" where
  "switch_and_nf c = nf_constr (constr_switch_left c)"


lemma interp_constr_impl_interp_switch_and_nf:
  "interp_constr ai vi (switch_and_nf c) = interp_constr ai vi c"
by (simp add: switch_and_nf_def)

lemma valid_constr_impl_valid_switch_constr:
  "valid_constr ai c = valid_constr ai (switch_and_nf c)"
by (simp add: valid_constr_def  interp_constr_impl_interp_switch_and_nf)


lemma no_right_switch_and_nf:
  "no_right_constr (switch_and_nf c)"
apply (case_tac c)
apply (simp add: switch_and_nf_def cnf_def)
done





lemma valid_no_right_eq:
  "\<lbrakk>valid_constr ai c;no_right_constr c;c= (Lesseq f g)\<rbrakk> \<Longrightarrow>
  interp_set_rep ai vi f = interp_set_rep ai vi g"
by ( simp add: valid_constr_def le_bot_eq)

fun 
  no_bconst_in_binop_constr :: "('n,'a) setconstr \<Rightarrow> bool"
where
  "no_bconst_in_binop_constr (Lesseq f g)= (no_bconst_in_binop f \<and> no_bconst_in_binop g)"






(*
export_code dnf cnf setconstr_to_list in OCaml file "setsolver.ml"
*)



end

