theory Aux imports Main

begin


subsection {* Isabelle notation *}


text {* Isabelle is based on a simply-typed lambda calculus. Function
definitions are introduced by keywords ``constdefs'' or ``primrec''
and are given in an ML-style syntax and should be directly
understandable.

Lemmas are introduced by the ``lemma'' keyword, followed by their
name, possibly some directives describing how they are used by
automated proof procedures. The notation @{text "\<lbrakk> P1, \<dots> Pn \<rbrakk> \<Longrightarrow> P"} is
a shorthand for an iterated implication @{text "P1 \<Longrightarrow> \<dots> Pn \<Longrightarrow> P"}.
Following the lemma are some proof commands (usually ``apply'' or
``by''). Often, these commands are not displayed in this text.

Some frequently used functions and predicates:
\begin{description}
\item[Option type and maps:]
  Partial functions from type @{text "'a"} to @{text "'b"} are modeled 
  by an @{text "'b option"} type, where the special value @{text None} 
  represents undefinedness and @{text "Some v"} the defined value @{text v}. 
  @{text "dom"} is the domain of a map (i.e. the values @{text "a"} which 
  map to @{text "Some"} value). 
  \item[Set functions:]
  The image of a set @{text A} under a function @{text f} is written as 
  @{text "f ` A"}. Lists are converted to sets by the function @{text set}.
\end{description}

*}

subsection {* Some list functions *}

lemma map_map_of_zip [rule_format]: 
  "length xs = length ys \<Longrightarrow> 
  distinct xs \<longrightarrow> map (map_of (zip xs ys)) xs = map Some ys"
apply (induct rule: list_induct2)
apply simp
apply clarify
apply (simp del: fun_upd_apply)
apply simp
done

lemma fst_set_zip: "
  length xs = length ys \<Longrightarrow> (fst ` set (zip xs ys)) = set xs"
by (induct rule: list_induct2) simp_all

lemma snd_set_zip: "
  length xs = length ys \<Longrightarrow> (snd ` set (zip xs ys)) = set ys"
by (induct rule: list_induct2) simp_all

lemma dom_option_map [simp]: "dom (\<lambda>x. Option.map f (s x)) = dom s"
by auto

lemma dom_if: "dom (\<lambda> x. if P x then f x else g x) = 
  (((dom f) \<inter> {x. P x}) \<union> ((dom g) \<inter> {x. \<not> P x}))"
by (simp add: dom_def Un_def ) fastforce

lemma image_dom: "inj f \<Longrightarrow>
  f ` dom g = dom (override_on empty (g \<circ> (inv f)) (range f))"
by (simp add: image_def) fastforce


subsection {* Maps and functions *}

text {* An alternative of the library lemma: *}

lemma map_of_map2:
  "map_of (map (\<lambda>(a, b). (a, f b)) xs) = (Option.map f) \<circ> (map_of xs)"
by (simp add: fun_eq_iff map_of_map )


lemma map_comp_map_upd_right [simp]: 
  "f o_m (g(a \<mapsto> b)) = (f o_m g)(a := f b)"
by (simp add: map_comp_def fun_eq_iff)


lemma map_comp_map_upd_left:
  "b \<notin> ran g \<Longrightarrow> f(b\<mapsto> a) o_m g = f o_m g"
apply (clarsimp simp add: map_comp_def fun_eq_iff split add: option.split)
apply (fastforce simp add: ran_def)
done

definition
  fun_upd_map :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a \<rightharpoonup> 'b) \<Rightarrow> 'a \<Rightarrow> 'b" where
  "fun_upd_map f m == \<lambda> x. case m x of None \<Rightarrow> f x | Some y \<Rightarrow> y"

lemma fun_upd_map_empty [simp]: "fun_upd_map f empty = f"
by (simp add: fun_upd_map_def)

lemma fun_upd_map_in_dom:
  "x \<in> dom m \<Longrightarrow> fun_upd_map f m x = the (m x)"
by (clarsimp simp add: fun_upd_map_def domIff)


lemma fun_upd_map_notin_dom:
  "x \<notin> dom m \<Longrightarrow> fun_upd_map f m x = f x"
by (clarsimp simp add: fun_upd_map_def domIff)

lemma fun_upd_map_upd: 
  "x \<notin> dom m \<Longrightarrow> fun_upd_map f (m(x\<mapsto>y)) = fun_upd_map (f(x:=y)) m"
by (auto simp add: fun_upd_map_def fun_eq_iff split add: option.split)

lemma fun_upd_map_assoc:
  "fun_upd_map (fun_upd_map f [xs [\<mapsto>] ys]) [x \<mapsto> y]
  = fun_upd_map f [xs [\<mapsto>] ys, x \<mapsto> y] "
by (simp add: fun_upd_map_def fun_eq_iff)


definition
  remap_fun :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a \<Rightarrow> 'b" where
  "remap_fun f xs ys == fun_upd_map f [xs[\<mapsto>](map f ys)]"



subsection {* Equality of functions *}

lemma ext_simp [simp]: "\<forall> x. f x = f' x \<Longrightarrow> f = f'"
by (simp add: fun_eq_iff)


lemma override_on_compl [simp]: 
  "override_on f g (- A) = override_on g f A"
by (simp add: override_on_def)

definition
  -- "f and f' possibly differ on A"
  funs_eq_except :: "['a \<Rightarrow> 'b, 'a \<Rightarrow> 'b, 'a set] \<Rightarrow> bool" where
  "funs_eq_except f f' A == \<forall> x. x \<notin> A \<longrightarrow> (f x) = (f' x)"

lemma funs_eq_except_fun_eq [simp]: "funs_eq_except f f A"
by (simp add: funs_eq_except_def)

lemma funs_eq_except_empty [simp]:
  "funs_eq_except f f' {} = (f = f')"
by (auto simp add: funs_eq_except_def fun_eq_iff)

lemma funs_eq_except_mono:
  "\<lbrakk> funs_eq_except f f' A; A \<subseteq> B \<rbrakk> \<Longrightarrow> funs_eq_except f f' B"
by (fastforce simp add: funs_eq_except_def)

lemma funs_eq_except_override_on:
  "A \<subseteq> B \<Longrightarrow> funs_eq_except f (override_on f f' A) B"
by (fastforce simp add: funs_eq_except_def override_on_def)


definition
  -- "f and f' are equal at least on A"
  funs_eq_on :: "['a \<Rightarrow> 'b, 'a \<Rightarrow> 'b, 'a set] \<Rightarrow> bool" where
  "funs_eq_on f f' A == \<forall> x \<in> A. (f x) = (f' x)"

lemma funs_eq_on_fun_refl [simp]: "funs_eq_on f f A"
by (simp add: funs_eq_on_def)

lemma funs_eq_on_sym: "funs_eq_on f f' A \<Longrightarrow> funs_eq_on f' f A"
by (simp add: funs_eq_on_def)

lemma funs_eq_on_trans: 
  "\<lbrakk> funs_eq_on f f' A; funs_eq_on f' f'' A\<rbrakk> \<Longrightarrow> funs_eq_on f f'' A"
by (simp add: funs_eq_on_def)

lemma funs_eq_on_UNIV [simp]:
  "funs_eq_on f f' UNIV = (f = f')"
by (auto simp add: funs_eq_on_def)

lemma funs_eq_on_empty [simp]:
  "funs_eq_on f f' {}"
by (simp add: funs_eq_on_def)

lemma funs_eq_on_mono:
  "\<lbrakk> funs_eq_on f f' A; B \<subseteq> A \<rbrakk> \<Longrightarrow> funs_eq_on f f' B"
by (fastforce simp add: funs_eq_on_def)


lemma funs_eq_on_override_on_split:
  "(funs_eq_on f (override_on f' f'' B) A) =
    (funs_eq_on f f'' (A \<inter> B) \<and> funs_eq_on f f' (A - B))"
by (fastforce simp add: funs_eq_on_def override_on_def)

lemma funs_eq_on_override_on1:
  "A \<subseteq> - B \<Longrightarrow> 
  (funs_eq_on f (override_on f' f'' B) A) = (funs_eq_on f f' A)"
by (fastforce simp add: funs_eq_on_def override_on_def)

lemma funs_eq_on_override_on2:
  "A \<subseteq> B \<Longrightarrow> 
  (funs_eq_on f (override_on f' f'' B) A) = (funs_eq_on f f'' A)"
by (fastforce simp add: funs_eq_on_def override_on_def)

lemma funs_eq_on_insert: 
  "funs_eq_on f f' (insert a A) = (f a = f' a \<and> funs_eq_on f f' A)"
by (simp add: funs_eq_on_def)

lemma funs_eq_on_union [simp]:
  "funs_eq_on f f' (A \<union> B) =
  (funs_eq_on f f' A \<and> funs_eq_on f f' B)"
by (fastforce simp add: funs_eq_on_def)

lemma funs_eq_on_remap_fun [rule_format]: 
  "length xs = length ys \<Longrightarrow> 
  funs_eq_on f f' A \<longrightarrow> set ys \<subseteq> A \<longrightarrow> distinct xs \<longrightarrow> 
  funs_eq_on (remap_fun f xs ys) (remap_fun f' xs ys) (A \<union> set xs)"
apply (induct rule: list_induct2)
apply (clarsimp simp add: remap_fun_def fun_upd_map_def funs_eq_on_def)+
done


subsection {* Logic *}

primrec
  conjs :: "bool list \<Rightarrow> bool" where
  "conjs [] = True"
| "conjs (x#xs) = (x \<and> conjs xs)"

lemma conjs_map: "(conjs (map P xs) = (\<forall> x \<in> set xs. P x))"
by (induct xs) auto

lemma biequation_left:
  "a = b \<Longrightarrow> ((c = a) = (c = b))"
by simp

lemma biequation_right:
  "a = b \<Longrightarrow> ((a = c) = (b = c))"
by simp

end
