theory Canonlang imports Interflang Errors


(* Datatypes for canonized expressions *)

begin

(* ---------------------------------------------------------------------- *)
subsection {* Datatypes for canonized expressions and statements *}


text {* The following datatypes make it easier to define recursive 
  functions over canonized expressions. 
 *}


datatype ('n,'v) canon_expr
  = CanVar 'n
  | CanPairE "(('n,'v) canon_expr)" "(('n,'v) canon_expr)"
  | CanFun funid  "(('n,'v) canon_expr)"

text {* Often (for example for compilation), it may be sufficient to have not
 completely canonized expressions, such as (CanWhen y e x), where e is a
 combinatorial expression.  In some occasions, however, that overly
 complicates matters (clock calculus). Therefore, combinatorial expressions
 may only occur in the form (CanCombin x e).
  *}

datatype ('n,'v,'tp) canon_stmt
  = CanEmptyStmt
  | CanCombin 'n "(('n,'v) canon_expr)"
  | CanPre 'n 'v 'n
  | CanWhen 'n 'n 'n
  | CanDefault 'n 'n 'n

  | CanPar "(('n,'v,'tp) canon_stmt)" "(('n,'v,'tp) canon_stmt)"
  | CanPCall "('n cname)" "('n list)"
  | CanLetv 'n 'tp "(('n,'v,'tp) canon_stmt)"

text {* All local variable declarations are raised to the outermost
  level of a procedure.
*}


  (* Full procedure definition, with header, parameter types and body *)
type_synonym ('n, 'tp) canon_tp_env = "'n \<Rightarrow> 'tp option"

datatype ('n,'v,'tp) canon_proc
  = CanProc  "('n proc_header)" "('n,'tp) var_types list" "(('n,'v,'tp) canon_stmt)"

primrec get_can_proc_header :: "('n,'v,'tp) canon_proc \<Rightarrow> ('n proc_header)"
where "get_can_proc_header (CanProc ph tps bd) = ph"
primrec get_can_proc_var_types :: "('n,'v,'tp) canon_proc \<Rightarrow> ('n,'tp) var_types list"
where "get_can_proc_var_types (CanProc ph tps bd) = tps"
primrec  get_can_proc_body :: "('n,'v,'tp) canon_proc \<Rightarrow> (('n,'v,'tp) canon_stmt)"
where "get_can_proc_body (CanProc ph tps bd) = bd"

  (* canon proc environment, both as assoc list 
  (result of parsing) and map (easier for proofs) *)
type_synonym ('n,'v,'tp) canon_proc_env 
  = "'n proc_header \<Rightarrow> ('n,'v,'tp) canon_proc option"

fun constr_canon_proc_env :: "('n,'v,'tp) canon_proc list \<Rightarrow> ('n,'v,'tp) canon_proc_env"
where
  "constr_canon_proc_env procs = map_of (map (\<lambda> pr. (get_can_proc_header pr, pr)) procs)"

datatype ('n,'v,'tp) canon_program
  = CanProgram "(('n,'v,'tp) canon_proc list)" "('n,'v,'tp) canon_stmt"

fun canon_proc_proj_interf
where "canon_proc_proj_interf (CanProc ph tps bd) = 
  (case ph of (ProcHeader pintn pimpl) \<Rightarrow> (Interf pintn tps))"

primrec
  fv_canon_expr :: "('n,'v) canon_expr \<Rightarrow> 'n list"
where
  "fv_canon_expr (CanVar x) = [x]"
| "fv_canon_expr (CanPairE e e') = 
        list_union (fv_canon_expr e) (fv_canon_expr e')"
| "fv_canon_expr (CanFun f e) = fv_canon_expr e"

primrec
  fv_canon_stmt :: "('n,'v,'tp) canon_stmt \<Rightarrow> 'n list"
where
  "fv_canon_stmt CanEmptyStmt = []"
| "fv_canon_stmt (CanCombin y e) = list_union [y] (fv_canon_expr e)"
| "fv_canon_stmt (CanPre y v x) = [y,x]"
| "fv_canon_stmt (CanWhen y x b) = [y,x,b]"
| "fv_canon_stmt (CanDefault y x z) = [y,x,z]"
| "fv_canon_stmt (CanPar c c') = list_union (fv_canon_stmt c) (fv_canon_stmt c')"
| "fv_canon_stmt (CanPCall cn args) = args"
| "fv_canon_stmt (CanLetv v clk c) = list_diff (fv_canon_stmt c) [v]"


fun var_types_names ::  "('n,'tp) var_types list \<Rightarrow> 'n list "
where "var_types_names xs = map get_var_types_name xs"

  (* rename tp to val_tp for value types *)
type_synonym 'n val_tp_env = "'n \<Rightarrow> val_tp option"


fun get_type_list ::  "('n,'tp) var_types list \<Rightarrow> 'tp list "
where "get_type_list xs = map get_var_types_type xs"


primrec tp_canon_expr :: "'n val_tp_env \<Rightarrow> ('n,'v) canon_expr \<Rightarrow> val_tp option"
where
  "tp_canon_expr tenv (CanVar n) = tenv n" |
  "tp_canon_expr tenv (CanPairE e1 e2) = 
  tp_pair (tp_canon_expr tenv e1) (tp_canon_expr tenv e2)" |
  "tp_canon_expr tenv (CanFun fid e) = tp_fun fid (tp_canon_expr tenv e)"


primrec
  wtpd_canon_stmt :: 
  "[('n, vals,  'c inter_typ) interf_env, ('n, val_tp) canon_tp_env, ('n, vals,  'c inter_typ) canon_stmt] \<Rightarrow> bool"
where
  "wtpd_canon_stmt ienv tenv CanEmptyStmt = True"
| "wtpd_canon_stmt ienv tenv (CanCombin x e) = 
    tp_compatible (tenv x) (tp_canon_expr tenv e)"
| "wtpd_canon_stmt ienv tenv (CanPre y v x) = 
    (tp_compatible (tenv y) (tenv x) \<and>  tp_compatible (tenv y) (Some (vals_tp v)))"
| "wtpd_canon_stmt ienv tenv (CanWhen y x b) = 
    (tp_compatible (tenv y) (tenv x) \<and> (tenv b) = (Some BoolT))"
| "wtpd_canon_stmt ienv tenv (CanDefault y x z) = 
    (tp_compatible (tenv y) (tenv x) \<and> tp_compatible (tenv y) (tenv z))"
| "wtpd_canon_stmt ienv tenv (CanPar c c') = 
     (wtpd_canon_stmt ienv tenv c \<and> wtpd_canon_stmt ienv tenv c')"
| "wtpd_canon_stmt ienv tenv (CanLetv v tp c) = 
     wtpd_canon_stmt ienv (tenv (v \<mapsto> get_inter_typ_valtp tp)) c"

| "wtpd_canon_stmt ienv tenv (CanPCall cn args)= 
  (case cn of Cname (ProcHeader ifn impn) cnum \<Rightarrow> 
      (case ienv ifn of
         Some (Interf ifn' tps) \<Rightarrow> (ifn' = ifn) 
             \<and> (list_inter args (var_types_names tps) = [])
             \<and> ((map (tenv) args) = map (Some \<circ> get_inter_typ_valtp) (get_type_list tps))
       | None \<Rightarrow> False ))"


definition
   type_env :: "('n,'tp) var_types list \<Rightarrow> ('n,'tp) canon_tp_env"
where  "type_env vars \<equiv> map_of (map (\<lambda> v. (get_var_types_name v, (get_var_types_type v))) vars )"

definition
   valtp_type_env :: "('n, 'c inter_typ) var_types list \<Rightarrow> ('n, val_tp) canon_tp_env"
where   "valtp_type_env vars \<equiv> 
  map_of (map (\<lambda> v. (get_var_types_name v,get_inter_typ_valtp (get_var_types_type v))) vars )"

fun val_tp_list::"('n, 'c inter_typ) var_types list \<Rightarrow>('n, val_tp) var_types list "
 where "val_tp_list tps =  
  map(\<lambda> tp. VarTypes (get_var_types_name tp) (get_inter_typ_valtp (get_var_types_type tp))) tps "

 fun  val_tp_interf::"('n, 'c inter_typ) interf \<Rightarrow> ('n, val_tp) interf"
where "val_tp_interf (Interf n tps) = Interf n (val_tp_list tps) "


primrec
  wtpd_canon_proc :: 
  "[('n, vals, 'c inter_typ) interf_env, ('n, vals, 'c inter_typ) canon_proc] \<Rightarrow> bool"
where
  "wtpd_canon_proc ienv (CanProc ph tpars bd) = 
  ( distinct (map get_var_types_name tpars)
  \<and> (wtpd_canon_stmt ienv (valtp_type_env tpars) bd))"

consts interf_env_conv ::
  "('n, 'v, 'c) canon_proc_env \<Rightarrow> ('n, 'v, 'c) interf_env"

definition
  wtpd_canon_proc_env :: "('n, vals, 'c inter_typ) canon_proc_env \<Rightarrow> bool"
where  "wtpd_canon_proc_env cpenv == 
  \<forall> pn \<in> dom cpenv. wtpd_canon_proc (interf_env_conv cpenv) (the (cpenv pn))"

definition
  wtpd_canon_proc_list ::
  "[('n, vals, 'c inter_typ) interf_env, ('n, vals, 'c inter_typ) canon_proc list] \<Rightarrow> bool"
where "wtpd_canon_proc_list ienv procs =
  (list_all (wtpd_canon_proc ienv) procs)"

lemma wtpd_canon_proc_env_list_equivs_aux [rule_format]:
  "distinct (map get_can_proc_header procs) \<longrightarrow> 
   (x \<in> set procs
  \<longrightarrow> (map_of (map (\<lambda>pr. (get_can_proc_header pr, pr)) procs) (get_can_proc_header x)) = Some x)"
apply (induct procs)
apply simp
apply clarsimp
apply (drule map_of_is_SomeD)
apply (auto )
done

lemma wtpd_canon_proc_env_distinct:
  "\<lbrakk> wtpd_canon_proc_env penv; penv pn = Some (CanProc pn tpars bd) \<rbrakk>
  \<Longrightarrow> distinct (var_types_names tpars)"
by (fastforce simp add: wtpd_canon_proc_env_def intro: domI dest: bspec)

(*
lemma wtpd_canon_proc_env_wtpd_canon_stmt:
  "\<lbrakk> wtpd_canon_proc_env penv; 
  penv pn = Some (CanProc pn tpars bd) \<rbrakk>
  \<Longrightarrow> wtpd_canon_stmt penv (valtp_type_env tpars) bd"
by (fastforce simp add: wtpd_canon_proc_env_def intro: domI dest: bspec)
*)


primrec
  canon_sub_expr :: "('n,'v) canon_expr \<Rightarrow> ('n,'v) canon_expr list"
where
  "canon_sub_expr (CanVar x) = [CanVar x]" |
  "canon_sub_expr (CanPairE e e') = 
        list_union 
          (list_union [(CanPairE e e')]
             (canon_sub_expr e)) (canon_sub_expr e')" |
  "canon_sub_expr (CanFun f e) = list_union [(CanFun f e)] (canon_sub_expr e)"


  (* -------------------------------------------------- *)
  (*  Lemmas: *)

lemma canon_sub_expr_self [simp]: "e \<in>set (canon_sub_expr e)"
by (induct e) auto

lemma fv_canon_expr_productive: "\<exists> v. v \<in> set (fv_canon_expr e)"
by (induct e) auto


lemma fv_canon_sub_expr [rule_format]:
  "\<forall>e'\<in>set (canon_sub_expr e). set (fv_canon_expr e') \<subseteq> set (fv_canon_expr e)"
by (induct e) auto


(* ---------------------------------------------------------------------- *)





(* ---------------------------------------------------------------------- *)
subsection {* Semantics of canonized expressions and statements *}


text {* Error semantics *}

primrec
  interp_canon_expr :: 
  "[('t::linorder, 'v::vals_cl, 've) error_fns, ('t,'n, 'v) trace, ('n,'v) canon_expr] 
  \<Rightarrow> ('t, 've) signal"
where
  "interp_canon_expr efs tr (CanVar x) = 
      apply_fun_pointwise (error_elem_lift efs) (trace_proj tr x)" |
  "interp_canon_expr efs tr (CanPairE e e') = 
       (error_pair efs
          (interp_canon_expr efs tr e)
          (interp_canon_expr efs tr e'))" |
  "interp_canon_expr efs tr (CanFun f e) = 
       apply_fun_pointwise 
           (error_fun_lift efs (interp_funid f))
           (interp_canon_expr efs tr e)"

inductive
  interp_canon_stmt :: 
     "[sync_error_handling,
       ('t::linorder,'n,'v::vals_cl) trace, ('n,'v,'tp) canon_proc_env, 
       ('t,'n,'v) trace \<Rightarrow> 'tp \<Rightarrow> 't set, ('n,'v,'tp) canon_stmt]
  \<Rightarrow> bool"
where
  interp_canonCanEmptyStmt [simp]: 
  "interp_canon_stmt efs tr penv tpi CanEmptyStmt"
| interp_canonCanCombin:
  "\<lbrakk> case efs of
       Blocking \<Rightarrow> error_free_signal
            (interp_canon_expr error_fns_lifted tr e)
     | Permissive \<Rightarrow> True;
  (trace_proj tr y = 
     interp_canon_expr error_fns_flat tr e) \<rbrakk>
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanCombin y e)"
| interp_canonCanPre:
  "trace_proj tr y = delay v (trace_proj tr x)
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanPre y v x)"
| interp_canonCanWhen:
  "trace_proj tr y = when (trace_proj tr x) (trace_proj tr b)
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanWhen y x b)"
| interp_canonCanDefault:
  "trace_proj tr y =default (trace_proj tr x) (trace_proj tr z)
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanDefault y x z)"
| interp_canonCanPar:
  "\<lbrakk>interp_canon_stmt efs tr penv tpi c; 
    interp_canon_stmt efs tr penv tpi c'\<rbrakk>
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanPar c c')"
| interp_canonCanPCall:
  "\<lbrakk> penv pn = Some (CanProc pn' tpars bd);
  interp_canon_stmt efs 
       (\<lambda>t. remap_fun (tr t) (var_types_names tpars) args)
       penv tpi bd \<rbrakk>
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi 
          (CanPCall (Cname pn cnum) args)"
| interp_canonCanLetv:
  "\<lbrakk> interp_canon_stmt efs (trace_upd tr v s) penv tpi c;
     dom s = tpi tr clk \<rbrakk>
  \<Longrightarrow> interp_canon_stmt efs tr penv tpi (CanLetv v clk c)"


primrec
  interp_canon_program :: "[sync_error_handling,
       ('t::linorder,'n,'v::vals_cl) trace, 
       ('t,'n,'v) trace \<Rightarrow> 'tp \<Rightarrow> 't set, ('n,'v,'tp) canon_program]
  \<Rightarrow> bool"
where
  "interp_canon_program efs tr tpi (CanProgram procs c) =
   interp_canon_stmt efs tr (constr_canon_proc_env procs) tpi c" 


lemma intp_canonCanCombin [simp]:
  "interp_canon_stmt efs tr penv tpi (CanCombin y e) =
  ((case efs of
       Blocking \<Rightarrow> error_free_signal
            (interp_canon_expr error_fns_lifted tr e)
     | Permissive \<Rightarrow> True) \<and>
   (trace_proj tr y = 
      interp_canon_expr error_fns_flat tr e))"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanCombin)


lemma intp_canonCanPre [simp]:
  "interp_canon_stmt efs tr penv tpi (CanPre y v x) =
  (trace_proj tr y = delay v (trace_proj tr x))"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanPre)

lemma intp_canonCanWhen [simp]:
  "interp_canon_stmt efs tr penv tpi (CanWhen y x b) =
  (trace_proj tr y = when (trace_proj tr x) (trace_proj tr b))"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanWhen)

lemma intp_canonCanDefault [simp]:
  "interp_canon_stmt efs tr penv tpi (CanDefault y x z) =
  (trace_proj tr y =default (trace_proj tr x) (trace_proj tr z))"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanDefault)

lemma intp_canonCanPar [simp]:
  "interp_canon_stmt efs tr penv tpi (CanPar c c') =
  (interp_canon_stmt efs tr penv tpi c \<and> 
   interp_canon_stmt efs tr penv tpi c')"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanPar)

lemma intp_canonCanPCall [simp]:
  "interp_canon_stmt efs tr penv tpi (CanPCall cn args) =
  (\<exists> pn cnum pn' tpars bd. 
    cn = (Cname pn cnum) \<and> 
    penv pn = Some (CanProc pn' tpars bd) \<and>
    interp_canon_stmt efs 
       (\<lambda>t. remap_fun (tr t) (var_types_names tpars) args)
       penv tpi bd)"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanPCall)

lemma intp_canonCanLetv [simp]:
  "interp_canon_stmt efs tr penv tpi (CanLetv v clk c) =
  (\<exists> s. interp_canon_stmt efs (trace_upd tr v s) penv tpi c \<and>
         dom s = tpi tr clk )"
by (auto elim: interp_canon_stmt.cases intro: interp_canonCanLetv)


end
