theory Errors imports Stream_fcts

begin



subsection {* Introducing error elements *}


datatype 'v error = Err | OK 'v

primrec
  cancel_err :: "'v error \<Rightarrow> 'v"
where
  "cancel_err (OK v) = v"

definition
  cancel_err_signal :: "('t, 'v error) signal \<Rightarrow> ('t, 'v) signal"
where  "cancel_err_signal s == (Option.map cancel_err) \<circ> s"

definition
  error_free_event :: "'v error option \<Rightarrow> bool"
where  "error_free_event e == (e \<noteq> (Some Err))"
definition  error_free_signal  :: "('t, 'v error) signal \<Rightarrow> bool"
where  "error_free_signal s == (\<forall> t. error_free_event (s t))"

lemma OK_cancel_err [simp, intro]:
  "\<lbrakk> error_free_event e; e = (Some v)\<rbrakk> \<Longrightarrow> OK (cancel_err v) = v"
by (case_tac v) (auto simp add: error_free_event_def)

lemma error_free_event_OK:
  "\<lbrakk> error_free_event e; e = Some x \<rbrakk> \<Longrightarrow> \<exists> v. x = OK v"
by (case_tac x) (auto simp add: error_free_event_def)

lemma error_free_event_lift_OK [simp]:
  "error_free_event (Option.map OK ev)"
by (simp add: Option.map_def error_free_event_def split add: option.split)

lemma error_free_signal_lift_OK [simp]:
  "error_free_signal (apply_fun_pointwise OK s)"
by (simp add: error_free_signal_def apply_fun_pointwise_def)


definition
  lift_err :: "[('v::vals_cl) error, 'v error] \<Rightarrow> 'v error"
where  "lift_err vo vo' == 
         (case vo of
            Err \<Rightarrow> Err
          | OK v \<Rightarrow> 
         (case vo' of
            Err \<Rightarrow> Err
          | OK v' \<Rightarrow> OK (pair_val (v, v'))))"

lemma [simp]: "lift_err Err x = Err"
by (simp add: lift_err_def)

lemma [simp]: "lift_err x Err = Err"
by (simp add: lift_err_def split add: error.split)

lemma [simp]: "lift_err (OK v) (OK v') = OK (pair_val (v, v'))"
by (simp add: lift_err_def split add: error.split)

definition
  lift_pair_err :: "[('v::vals_cl) error option, 'v error option] \<Rightarrow> 'v error option"
where  "lift_pair_err vo vo' == 
         (case vo of
            None \<Rightarrow> 
            (case vo' of
               None \<Rightarrow> None
             | Some v' \<Rightarrow> Some Err)
          | Some v \<Rightarrow> 
            (case vo' of
               None \<Rightarrow> Some Err
             | Some v' \<Rightarrow> Some (lift_err v v')))"


primrec
  lift_fun_error :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a error \<Rightarrow> 'b error)"
where
  "lift_fun_error f Err = Err"|
  "lift_fun_error f (OK v) = OK (f v)"

lemma lift_fun_error_inv_Err [simp]: 
  "(lift_fun_error f x = Err) = (x = Err)"
by (case_tac x) auto


definition
  lift_signals_error :: 
  "[('t,'v) signal \<Rightarrow> ('t,'v) signal \<Rightarrow> ('t,'v) signal, 
    ('t,'v error) signal, ('t,'v error) signal] 
  \<Rightarrow> ('t,'v error) signal"
where  "lift_signals_error ff s s' == (\<lambda> t. 
    if ((error_free_event (s t)) \<and> (error_free_event (s' t)))
    then apply_fun_pointwise OK (ff (cancel_err_signal s) (cancel_err_signal s')) t
    else (Some Err))"


lemma lift_pair_err_None: "(lift_pair_err v v' = None) = (v = None \<and> v' = None)"
by (auto 
  simp add: lift_pair_err_def lift_err_def proj1_pair_val proj2_pair_val 
  split add: option.split error.split)


  (* ------------------------------------------------------------ *)
  (* Stream functions with errors *)

text {* The function delay_error has been introduced to keep track of error elements present in a stream, necessary for proving the lemma error_free_signal_delay_imp_error_free. This is not achieved by function delay alone: 
If a stream is finite, the last element of s can be an error element, which does 
  not appear in (delay (OK v) s).
*}

definition
  signalpair_error ::
     "[('t, ('v::vals_cl) error) signal, ('t,'v error) signal] \<Rightarrow> ('t,'v error) signal"
where  "signalpair_error s s' == (\<lambda> t. (lift_pair_err (s t) (s' t)))"

definition
  delay_error :: "'v \<Rightarrow> ('t::linorder,'v error) signal \<Rightarrow> ('t,'v error) signal"
where  "delay_error v s == (delay (if (error_free_signal s) then (OK v) else Err) s)"

  (* ------------------------------------------------------------ *)
  (* Error function records *)

text {* Error function record: expression interpretation (with errors)
  is parameterized by elements of this record *}

record ('t, 'v, 've) error_fns = 
  error_elem_lift :: "'v \<Rightarrow> 've"
  error_fun_lift :: "('v \<Rightarrow> 'v) \<Rightarrow> ('ve \<Rightarrow> 've)"
  error_combine :: "(('t,'v) signal \<Rightarrow> ('t,'v) signal \<Rightarrow> ('t,'v) signal) \<Rightarrow>
  (('t,'ve) signal \<Rightarrow> ('t,'ve) signal \<Rightarrow> ('t,'ve) signal)"
  error_pair :: "[('t,'ve) signal, ('t,'ve) signal] \<Rightarrow> ('t,'ve) signal"
  error_delay :: "['v, ('t,'ve) signal] \<Rightarrow> ('t,'ve) signal"


text {* Error interpretation that ignores errors *}

datatype sync_error_handling = Blocking | Permissive

definition
  error_fns_flat :: "('t::linorder, ('v::vals_cl), 'v) error_fns"
where  "error_fns_flat == 
     \<lparr> error_elem_lift = id, 
       error_fun_lift = id, 
       error_combine = id,
       error_pair = signalpair,
       error_delay = delay \<rparr>"


lemma [simp]: "error_elem_lift error_fns_flat = id"
by (simp add: error_fns_flat_def)
lemma [simp]: "error_fun_lift error_fns_flat = id"
by (simp add: error_fns_flat_def)
lemma [simp]: "error_combine error_fns_flat = id"
by (simp add: error_fns_flat_def)
lemma [simp]: "error_pair error_fns_flat = signalpair"
by (simp add: error_fns_flat_def)
lemma [simp]: "error_delay error_fns_flat = delay"
by (simp add: error_fns_flat_def)


text {* Error interpretation that propagates errors *}
definition
  error_fns_lifted :: "('t::linorder, ('v::vals_cl), 'v error) error_fns"
where  "error_fns_lifted == 
     \<lparr> error_elem_lift = OK, 
       error_fun_lift = lift_fun_error, 
       error_combine = lift_signals_error,
       error_pair = signalpair_error,
       error_delay = delay_error \<rparr>"

lemma [simp]: "error_elem_lift error_fns_lifted = OK"
by (simp add: error_fns_lifted_def)
lemma [simp]: "error_fun_lift error_fns_lifted = lift_fun_error"
by (simp add: error_fns_lifted_def)
lemma [simp]: "error_combine error_fns_lifted = lift_signals_error"
by (simp add: error_fns_lifted_def)
lemma [simp]: "error_pair error_fns_lifted = signalpair_error"
by (simp add: error_fns_lifted_def)
lemma [simp]: "error_delay error_fns_lifted = delay_error"
by (simp add: error_fns_lifted_def)


  (* ---------------------------------------------------------------------- *)


text {* Propagation of error-freeness and Cancellation of errors in a signal *}
  


lemma error_free_signal_lift_signals_error [simp]:
  "error_free_signal (lift_signals_error ff s s') = 
  (error_free_signal s \<and> error_free_signal s')"
apply (clarsimp simp add: error_free_signal_def lift_signals_error_def)
apply (simp add: apply_fun_pointwise_def)
apply (fastforce simp add: Option.map_def error_free_event_def
       split add: option.split)
done

lemma [simp]: "lift_pair_err (Some Err) x = Some Err"
by (simp add: lift_pair_err_def split add: option.split)

lemma [simp]: "lift_pair_err x (Some Err) = Some Err"
by (simp add: lift_pair_err_def split add: option.split)

lemma error_free_signal_signalpair_error [rule_format]:
  "error_free_signal (signalpair_error s s') \<longrightarrow>
  (error_free_signal s \<and> error_free_signal s')"
apply (clarsimp simp add: error_free_signal_def signalpair_error_def)
apply (simp add: error_free_event_def)
apply (rule conjI)
apply clarsimp apply (drule_tac x=t in spec) apply simp
apply clarsimp apply (drule_tac x=t in spec) apply simp
done



lemma error_free_event_lift [simp]:
  "error_free_event (apply_fun_pointwise (lift_fun_error f) s t)
  = error_free_event (s t)"
by (simp add: apply_fun_pointwise_def error_free_event_def)

lemma error_free_signal_lift [simp]:
  "error_free_signal (apply_fun_pointwise (lift_fun_error f) s)
   = error_free_signal s"
by (simp add: error_free_signal_def)


lemma previous_defined_result [rule_format]:
  "discrete_event_signal s
  \<Longrightarrow> (previous_defined v s t = x) \<longrightarrow> ((v = x) \<or> (\<exists> t'. t' < t \<and> s t' = Some x))"
apply (clarsimp simp add: previous_defined_def Let_def)
apply (subgoal_tac "finite {t' \<in> dom s. t' < t}")
apply (fastforce dest: Max_in)
apply (drule_tac t=t in discrete_event_signal_finite_dom)
apply (rule finite_subset)
prefer 2
apply fastforce+
done

lemma error_free_imp_error_free_signal_delay [rule_format]:
  "discrete_event_signal s
  \<Longrightarrow> error_free_signal s \<longrightarrow> error_free_signal (delay_error v s)"
apply (simp add: error_free_signal_def)
apply (simp add: error_free_event_def)
apply (clarsimp simp add: delay_error_def delay_def split add: option.split)
apply (rule conjI)
apply clarsimp

apply (fast dest: previous_defined_result)
apply (clarsimp simp add: error_free_signal_def error_free_event_def)
done


lemma error_free_event_lift_pair_err: 
  "\<lbrakk>error_free_event ev; error_free_event ev';
        (\<exists>y. ev' = Some y) = (\<exists>y. ev = Some y) \<rbrakk>
        \<Longrightarrow> error_free_event (lift_pair_err ev ev')"
by (simp add: error_free_event_def lift_pair_err_def split add: option.split)
   (auto simp add: lift_err_def split add: error.split_asm)

lemma error_free_signal_sync_signals:
  "\<lbrakk> error_free_signal s; error_free_signal s'; sync_signals {s, s'} \<rbrakk>
  \<Longrightarrow> error_free_signal (signalpair_error s s')"
apply (clarsimp simp add: error_free_signal_def signalpair_error_def)
apply (clarsimp simp add: sync_signals_def dom_def set_eq_iff)
apply (fast intro: error_free_event_lift_pair_err)
done


lemma previous_defined_Min: "\<lbrakk> discrete_event_signal s; s t = Some y \<rbrakk> \<Longrightarrow>
     previous_defined v s (Min {t' \<in> dom s. t' \<le> t}) = v"
apply (clarsimp simp add: previous_defined_def Let_def)
apply (rename_tac x z)
apply (drule_tac t=t in discrete_event_signal_finite_dom)
apply (subgoal_tac "x \<notin> {t' \<in> dom s. t' \<le> t}")
  prefer 2 apply (fastforce dest: Min_le)
apply (clarsimp simp add: domI)
apply (frule Min_in) apply fastforce
apply clarsimp
apply order
done

lemma error_free_event_Err:
  "\<lbrakk> discrete_event_signal s;
  \<forall> t'. error_free_event (delay Err s t') \<rbrakk> \<Longrightarrow>
   error_free_event (s t)"
apply (clarsimp simp add: error_free_event_def)
apply (clarsimp simp add: delay_def split add: option.split_asm)
apply (drule_tac x="Min {t' \<in> dom s. t' \<le> t}" in spec)
apply (subgoal_tac "finite {t' \<in> dom s. t' \<le> t}")
  prefer 2 apply simp
apply (fastforce dest: Min_in simp add: previous_defined_Min)
done


lemma error_free_signal_delay_imp_error_free [rule_format]:
  "discrete_event_signal s \<Longrightarrow> 
  error_free_signal (delay_error v s) \<longrightarrow> error_free_signal s"
by (clarsimp simp add: error_free_signal_def delay_error_def
  split add: split_if_asm)
   (fast intro: error_free_event_Err)

lemma error_free_signal_delay_error [simp]:
  "discrete_event_signal s \<Longrightarrow>
  error_free_signal (delay_error v s) = error_free_signal s"
apply (rule iffI)
apply (fast intro: error_free_signal_delay_imp_error_free)
apply (fast intro: error_free_imp_error_free_signal_delay)
done

lemma cancel_err_lift_err [simp]:
  "lift_err a b \<noteq> Err \<Longrightarrow>
           cancel_err (lift_err a b) = pair_val (cancel_err a, cancel_err b)"
by (fastforce simp add: lift_err_def split add: error.split)


lemma lift_OK_cancel_err [simp]:
  "error_free_signal s \<Longrightarrow> 
  apply_fun_pointwise OK (cancel_err_signal s) = s"
by (auto simp add: error_free_signal_def
    cancel_err_signal_def 
    apply_fun_pointwise_def fun_eq_iff Option.map_def
    split add: option.split)

lemma cancel_err_lift_OK [simp]:
  "cancel_err_signal (apply_fun_pointwise OK s) = s"
by (simp add: cancel_err_signal_def 
    apply_fun_pointwise_def comp_def fun_eq_iff Option.map_def
    split add: option.split)


lemma cancel_lift_signals_error [simp]:
  "\<lbrakk> error_free_signal s; error_free_signal s' \<rbrakk>
 \<Longrightarrow> cancel_err_signal (lift_signals_error ff s s') = 
   (ff (cancel_err_signal s) (cancel_err_signal  s'))"
by (simp add: cancel_err_signal_def lift_signals_error_def 
  Option.map_def comp_def error_free_signal_def
  fun_eq_iff apply_fun_pointwise_def error_free_event_def
  split add: option.split)


lemma cancel_lift_pair_err: "\<lbrakk>  error_free_event (lift_pair_err e e') \<rbrakk> \<Longrightarrow>
  Option.map cancel_err (lift_pair_err e e') =
           lift_pair (Option.map cancel_err e) (Option.map cancel_err e')"
apply (simp add: Option.map_def lift_pair_def lift_pair_err_def 
  split add: option.split)
apply (fastforce simp add: error_free_event_def split add: option.split_asm)
done

lemma cancel_signalpair_error [simp]:
  "error_free_signal (signalpair_error s s')
 \<Longrightarrow> cancel_err_signal (signalpair_error s s') = 
   (signalpair (cancel_err_signal s) (cancel_err_signal  s'))"
apply (clarsimp simp add: cancel_err_signal_def error_free_signal_def)
apply (clarsimp simp add: signalpair_error_def signalpair_def comp_def fun_eq_iff)
apply (fastforce simp add: cancel_lift_pair_err)
done


lemma cancel_apply_fun_pointwise [simp]:
  "error_free_signal s
 \<Longrightarrow> cancel_err_signal (apply_fun_pointwise (lift_fun_error f) s) = 
  (apply_fun_pointwise f (cancel_err_signal s))"
apply (simp add: cancel_err_signal_def 
  Option.map_def comp_def error_free_signal_def
  fun_eq_iff apply_fun_pointwise_def 
  split add: option.split)
apply (fastforce dest: error_free_event_OK)
done

lemma cancel_err_previous_defined: "
  discrete_event_signal s \<Longrightarrow>
  cancel_err (previous_defined (OK v) s t) =
  previous_defined v (\<lambda> x. Option.map cancel_err (s x)) t"
apply (clarsimp simp add: previous_defined_def Let_def)
apply (rule conjI)
apply fastforce
apply (clarsimp simp add: Option.map_def split add: option.split)
apply (subgoal_tac "finite {t' \<in> dom s. t' < t}")
apply (fastforce dest: Max_in) 
apply fastforce
done

lemma cancel_delay_error [simp]:
  "\<lbrakk> discrete_event_signal s; error_free_signal s \<rbrakk> \<Longrightarrow> 
  cancel_err_signal (delay_error v s) = 
  (delay v (cancel_err_signal s))"
by (clarsimp simp add: cancel_err_signal_def
  Option.map_def comp_def
  delay_error_def fun_eq_iff delay_def
  split add: option.split)
  (simp add: cancel_err_previous_defined Option.map_def)

lemma dom_delay_error [simp]: "dom (delay_error v s) = dom s"
by (simp add: delay_error_def)

lemma discrete_event_signal_delay_error [simp]:
  "discrete_event_signal s 
  \<Longrightarrow> discrete_event_signal (delay_error v s)"
by (clarsimp simp add: discrete_event_signal_def)

lemma dom_signalpair_error [simp]:
  "dom (signalpair_error s s') = (dom s \<union> dom s') "
apply (simp add: dom_def signalpair_error_def)
apply (rule set_eqI)
apply (simp only: not_None_eq [THEN sym])
apply (simp add: lift_pair_err_None)
done

lemma discrete_event_signal_signalpair_error [simp]:
  "\<lbrakk> discrete_event_signal s; discrete_event_signal s' \<rbrakk>
  \<Longrightarrow> discrete_event_signal (signalpair_error s s')"
by (clarsimp simp add: discrete_event_signal_def)
   (fastforce simp only: Int_Un_distrib)


lemma dom_cancel_err_signal [simp]:
  "dom (cancel_err_signal s) = dom s"
by (simp add: cancel_err_signal_def Option.map_def comp_def dom_def)
   (auto split add: option.split_asm)

lemma discrete_event_signal_cancel_err_signal [simp]:
  "discrete_event_signal s
  \<Longrightarrow> discrete_event_signal (cancel_err_signal s)"
by (clarsimp simp add: discrete_event_signal_def)


lemma error_free_event_Some1:
  "\<lbrakk> error_free_event (lift_pair_err (s t) (s' t)); s t = Some y\<rbrakk>
          \<Longrightarrow> \<exists>y. s' t = Some y"
by (fastforce simp add: error_free_event_def lift_pair_err_def
  split add:  option.split_asm)

lemma error_free_event_Some2:
  "\<lbrakk> error_free_event (lift_pair_err (s t) (s' t)); s' t = Some y\<rbrakk>
          \<Longrightarrow> \<exists>y. s t = Some y"
by (fastforce simp add: error_free_event_def lift_pair_err_def
  split add:  option.split_asm)

lemma error_free_signal_signalpair_dom:
  "error_free_signal (signalpair_error s s') \<Longrightarrow>
  (dom s = dom s')"
by (simp add: error_free_signal_def signalpair_error_def dom_def)
 (fastforce intro: error_free_event_Some1 error_free_event_Some2)

end

