theory Setsolver imports Main

begin

  (* set expressions, not containing set difference / complement *)
datatype setop = SInter | SUnion

datatype ('n, 'c) setexpr
  = SVar 'n
  | SConst 'c
  | SEmpty
  | SUniv
  | SOp setop "(('n, 'c) setexpr)" "(('n, 'c) setexpr)"


  (* set constraints, representing se1 \<subseteq> se2 *)
datatype ('n, 'c) setconstr 
  = SCTrue
  | SCSubseteq  "(('n, 'c) setexpr)" "(('n, 'c) setexpr)"
  | SCSeteq  "(('n, 'c) setexpr)" "(('n, 'c) setexpr)"


primrec
  interp_setop :: "setop \<Rightarrow> 's set \<Rightarrow> 's set \<Rightarrow> 's set"
where
  "interp_setop SInter = (op Int)" |
  "interp_setop SUnion = (op Un)"

primrec
  interp_setexpr :: "['n \<Rightarrow> 's set, 'c \<Rightarrow> 's set, ('n, 'c) setexpr] \<Rightarrow> 's set"
where
  "interp_setexpr iv ic (SVar n) = iv n"|
  "interp_setexpr iv ic (SConst c) = ic c"|
  "interp_setexpr iv ic SEmpty = {}"|
  "interp_setexpr iv ic SUniv = UNIV"|
  "interp_setexpr iv ic (SOp sop e e') = 
     (interp_setop sop) (interp_setexpr iv ic e) (interp_setexpr iv ic e')"

primrec
  interp_setconstr :: "['n \<Rightarrow> 's set, 'c \<Rightarrow> 's set, ('n, 'c) setconstr] \<Rightarrow> bool"
where
  "interp_setconstr iv ic SCTrue = True"|
  "interp_setconstr iv ic (SCSubseteq e e') = 
       ((interp_setexpr iv ic e) \<subseteq> (interp_setexpr iv ic e'))"|
  "interp_setconstr iv ic (SCSeteq e e') = 
       ((interp_setexpr iv ic e) = (interp_setexpr iv ic e'))"

definition
  valid_setconstr :: "['c \<Rightarrow> 's set, ('n, 'c) setconstr] \<Rightarrow> bool"
where  "valid_setconstr ic cnstr == \<forall> iv. interp_setconstr iv ic cnstr"

fun atomic :: "('n, 'c) setexpr \<Rightarrow> bool"
where 
 "atomic (SOp sop e e') = False"
| "atomic _ = True"

  (* clausalform sop e if e is of the form (modulo associativity):
  a1 sop a2 sop .. sop an,   with ai atomic  *)
fun
  clausalform_sop :: "[setop, ('n, 'c) setexpr] \<Rightarrow> bool"
where
  "clausalform_sop sop (SOp sop' e e') =
    ((sop = sop') \<and> clausalform_sop sop e \<and> clausalform_sop sop e')"
| "clausalform_sop sop _ = True"

  (* 
  transforms 
  e1 \<inter> \<dots> \<inter> en into [e1 \<dots> en]
  respectively
  e1 \<union> \<dots> \<union> en into [e1 \<dots> en]
 *)
fun
  sop_decompose :: "[setop, ('n, 'c) setexpr] \<Rightarrow> ('n, 'c) setexpr list"
where
  "sop_decompose sop (SOp sop' e e') = 
     (if sop = sop'
      then (sop_decompose sop e @ sop_decompose sop e')
      else [(SOp sop' e e')])"
| "sop_decompose sop x = [x]"


lemma sop_decompose_atomic [rule_format]: 
  "clausalform_sop sop e \<longrightarrow> (\<forall> a\<in> set (sop_decompose sop e). atomic a)"
by (induct e) auto

  (* the inverse of sop_decompose:
  turns [e1 \<dots> en] into e1 \<inter> \<dots> \<inter> en respectively e1 \<union> \<dots> \<union> en
 *)
primrec
  sop_compose :: "[setop, ('n, 'c) setexpr list] \<Rightarrow> ('n, 'c) setexpr"
where
  "sop_compose sop [] = (case sop of SInter \<Rightarrow> SUniv | SUnion \<Rightarrow> SEmpty)"|
  "sop_compose sop (x # xs) = (SOp sop x (sop_compose sop xs))"


lemma sop_compose_append_SInter [simp]:
  "interp_setexpr iv ic (sop_compose SInter (xs @ ys)) = 
  (interp_setexpr iv ic (sop_compose SInter xs)) \<inter> 
  (interp_setexpr iv ic (sop_compose SInter ys))"
by (induct xs) auto


lemma sop_compose_append_SUnion [simp]:
  "interp_setexpr iv ic (sop_compose SUnion (xs @ ys)) = 
  (interp_setexpr iv ic (sop_compose SUnion xs)) \<union> 
  (interp_setexpr iv ic (sop_compose SUnion ys))"
by (induct xs) auto


lemma sop_decompose_compose [simp]:
  "interp_setexpr iv ic (sop_compose sop (sop_decompose sop e))
  = interp_setexpr iv ic e"
apply (induct e)
apply (simp split add: setop.split)+
apply (rule conjI)
apply clarsimp+
done

lemma clausalform_Inter [rule_format]:
  "clausalform_sop SInter e \<longrightarrow>
  (\<Inter>a\<in> set (sop_decompose sop e). interp_setexpr iv ic a) = interp_setexpr iv ic e"
by (induct e) auto

lemma clausalform_Union [rule_format]:
  "clausalform_sop SUnion e \<longrightarrow>
  (\<Union>a\<in> set (sop_decompose sop e). interp_setexpr iv ic a) = interp_setexpr iv ic e"
by (induct e) auto


(*
fun atomic_union :: "('n, 'c) setexpr \<Rightarrow> bool"
where 
  "atomic_union (SVar n) = True"
| "atomic_union (SConst c) = True"
| "atomic_union (SUnion e e') = (atomic e \<and> atomic_union e')"
| "atomic_union _ = False"
*)

primrec
  nf_setconstr :: "('n, 'c) setconstr \<Rightarrow> bool"
where
  "nf_setconstr SCTrue = True"|
  "nf_setconstr (SCSubseteq e e') = (clausalform_sop SInter e \<and> clausalform_sop SUnion e')"|
  "nf_setconstr (SCSeteq e e') = False"

definition
  simplify_inter_SEmpty :: "('n, 'c) setexpr list \<Rightarrow> ('n, 'c) setexpr list"
where  "simplify_inter_SEmpty decs == if (SEmpty \<in>set decs) then [SEmpty] else decs"

definition
  simplify_inter_SUniv :: "('n, 'c) setexpr list \<Rightarrow> ('n, 'c) setexpr list"
where  "simplify_inter_SUniv == filter (\<lambda> a. a \<noteq> SUniv)"

definition
  simplify_union_SEmpty :: "('n, 'c) setexpr list \<Rightarrow> ('n, 'c) setexpr list"
where  "simplify_union_SEmpty == filter (\<lambda> a. a \<noteq> SEmpty)"

definition
  simplify_union_SUniv :: "('n, 'c) setexpr list \<Rightarrow> ('n, 'c) setexpr list"
where  "simplify_union_SUniv decs == if (SUniv \<in>set decs) then [SUniv] else decs"


(*lemma decomp_inter_SEmpty:
  "(\<Inter>a\<in> set decs. interp_setexpr iv ic a) = 
   (\<Inter>a\<in> set (simplify_inter_SEmpty decs). interp_setexpr iv ic a)"
apply  (induct decs)
apply simp
apply (simp add: simplify_inter_SEmpty_def)
apply (simp (no_asm) add: simplify_inter_SEmpty_def)
done*)

lemma decomp_inter_SUniv:
  "(\<Inter>a\<in> set decs. interp_setexpr iv ic a) = 
   (\<Inter>a\<in> set (simplify_inter_SUniv decs). interp_setexpr iv ic a)"
by (fastforce simp add: simplify_inter_SUniv_def)

lemma decomp_union_SEmpty:
  "(\<Union>a\<in> set decs. interp_setexpr iv ic a) = 
   (\<Union>a\<in> set (simplify_union_SEmpty decs). interp_setexpr iv ic a)"
by (fastforce simp add: simplify_union_SEmpty_def)

(*lemma decomp_union_SUniv:
  "(\<Union>a\<in> set decs. interp_setexpr iv ic a) = 
   (\<Union>a\<in> set (simplify_union_SUniv decs). interp_setexpr iv ic a)"
apply  (induct decs)
apply (simp add: simplify_union_SUniv_def)
apply simp
apply (simp (no_asm) add: simplify_union_SUniv_def)
done*)


datatype ('n, 'c) subsetconstr_decomp 
  = SDecomp "(('n, 'c) setexpr list)" "(('n, 'c) setexpr list)" 
            "(('n, 'c) setexpr list)" "(('n, 'c) setexpr list)"

definition
  valid_subsetconstr_decomp :: 
  "['c \<Rightarrow> 's set, ('n, 'c) subsetconstr_decomp] \<Rightarrow> bool"
where  "valid_subsetconstr_decomp ic sdec == 
      (case sdec of
       SDecomp inters_v inters_nv unions_v unions_nv \<Rightarrow>
         (\<forall> iv. 
          (\<Inter>a\<in>(set inters_v \<union> set inters_nv). (interp_setexpr iv ic a)) 
        \<subseteq> (\<Union>a\<in>(set unions_v \<union> set unions_nv). (interp_setexpr iv ic a))))"

(* To be redone, with other data structure
lemma "\<lbrakk> clausalform_sop SInter e; clausalform_sop SUnion e'\<rbrakk> \<Longrightarrow> 
  valid_setconstr ic (SCSubseteq e e') = 
  valid_subsetconstr_decomp ic ((sop_decompose SInter e), (sop_decompose SUnion e'))"
by (simp add: valid_setconstr_def valid_subsetconstr_decomp_def
    clausalform_Inter clausalform_Union)
*)

(* still messy
lemma "valid_subsetconstr_decomp ic (a # inters, unions)"
lemma 
"(\<Inter>a\<in> set (insert x decs). interp_setexpr iv ic a)
= (\<Inter>a\<in> set decs. interp_setexpr iv ic a)"
*)

lemma "SVar n \<in> set unions_v \<Longrightarrow>
  valid_subsetconstr_decomp ic 
      (SDecomp (SVar n # inters_v) inters_nv unions_v unions_nv) =
  valid_subsetconstr_decomp ic (SDecomp [] [SEmpty] [] [SUniv])"
by (fastforce simp add: valid_subsetconstr_decomp_def)



(*
lemma "(\<forall> (i:: 'a \<Rightarrow> 'b \<Rightarrow> bool). i(a) \<inter> i(b) \<subseteq> i(c)) = (\<forall> (i:: 'a \<Rightarrow> 'b \<Rightarrow> bool). i(b) \<subseteq> i(c))"
apply (rule iffI)
prefer 2
apply clarsimp apply (drule_tac x=i in spec) apply fast
apply clarsimp apply (drule_tac x=
*)

primrec
  fv :: "('n, 'c) setexpr \<Rightarrow> 'n set"
where
  "fv (SVar n) = {n}"|
  "fv (SConst c) = {}"|
  "fv SEmpty = {}"|
  "fv SUniv = {}"|
  "fv (SOp sop e e') = fv e \<union> fv e'"


lemma interp_setexpr_upd [rule_format, simp]: 
  "n \<notin> fv e \<longrightarrow>
  interp_setexpr (iv(n := S)) ic e = interp_setexpr iv ic e"
by (induct e) auto



lemma "n \<notin> \<Union> fv ` ((set inters_v) \<union> (set inters_nv) 
                    \<union> (set unions_v) \<union> (set unions_nv)) \<Longrightarrow>
  valid_subsetconstr_decomp ic 
      (SDecomp ((SVar n) # inters_v) inters_nv unions_v unions_nv) =
  valid_subsetconstr_decomp ic 
      (SDecomp (inters_v) inters_nv unions_v unions_nv)"
apply (rule iffI)
apply (clarsimp simp add: valid_subsetconstr_decomp_def)
apply (drule_tac x="iv(n:=UNIV)" in spec)
apply (clarsimp simp add: INT_Un subset_eq)
apply (subgoal_tac "x \<in> (\<Inter>i\<in>set inters_v. interp_setexpr iv ic i) \<inter>
            (\<Inter>i\<in>set inters_nv. interp_setexpr iv ic i)")
  prefer 2 apply (simp add: ball_Un)
apply (drule bspec, assumption)
apply simp

apply (clarsimp simp add: valid_subsetconstr_decomp_def)
apply (drule_tac x="iv" in spec)
apply (subgoal_tac "x \<in> (\<Inter>i\<in>set inters_v \<union> set inters_nv. interp_setexpr iv ic i)")
  prefer 2 apply (simp add: ball_Un)
apply (drule subsetD) apply assumption
apply simp
done

lemma "n \<notin> \<Union> fv ` ((set inters_v) \<union> (set inters_nv) 
                    \<union> (set unions_v) \<union> (set unions_nv)) \<Longrightarrow>
  valid_subsetconstr_decomp ic 
      (SDecomp inters_v inters_nv (SVar n # unions_v) unions_nv) =
  valid_subsetconstr_decomp ic 
      (SDecomp inters_v inters_nv unions_v unions_nv)"
apply (rule iffI)
apply (clarsimp simp add: valid_subsetconstr_decomp_def)
apply (drule_tac x="iv(n:={})" in spec)
apply (clarsimp simp add: INT_Un subset_eq)
apply (subgoal_tac "x \<in> (\<Inter>i\<in>set inters_v. interp_setexpr iv ic i) \<inter>
            (\<Inter>i\<in>set inters_nv. interp_setexpr iv ic i)")
  prefer 2 apply (simp add: ball_Un)
apply (drule bspec, assumption)
apply simp

apply (clarsimp simp add: valid_subsetconstr_decomp_def)
apply (drule_tac x="iv" in spec)
apply (subgoal_tac "x \<in> (\<Inter>i\<in>set inters_v \<union> set inters_nv. interp_setexpr iv ic i)")
  prefer 2 apply (simp add: ball_Un)
apply (drule subsetD) apply assumption
apply simp
done


lemma "valid_subsetconstr_decomp ic 
      (SDecomp (remdups inters_v) inters_nv (remdups unions_v) unions_nv) =
  valid_subsetconstr_decomp ic 
      (SDecomp inters_v inters_nv unions_v unions_nv)"
by (clarsimp simp add: valid_subsetconstr_decomp_def)

end
