(* Functions for combining streams *)

theory Stream_fcts imports Streams Values

begin

  (* ---------------------------------------------------------------------- *)
subsection {* Prerequisites *}

text {*
For the time being, the value domain (type parameter @{text "'v"}) remains abstract. 
There are just some abstract functions for testing properties of values:

A test wheter a value is true (true_val) and a pair constructors and destructors.
 *}


  (* ---------------------------------------------------------------------- *)

subsection {* Semantics of expressions  - auxiliary functions *}

definition
  lift_pair :: "[('v::vals_cl) option, 'v option] \<Rightarrow> 'v option" where
  "lift_pair vo vo' == 
         (case vo of
            None \<Rightarrow> None
          | Some v \<Rightarrow> 
         (case vo' of
            None \<Rightarrow> None
          | Some v' \<Rightarrow> Some (pair_val (v, v'))))"

definition
  apply_fun_pointwise :: "['a \<Rightarrow> 'b, ('t,'a) signal] \<Rightarrow> ('t,'b) signal" where
  "apply_fun_pointwise f  ==  \<lambda> s.(Option.map f) \<circ> s"

lemma apply_fun_pointwise_id [simp]: "apply_fun_pointwise id = id"
by (clarsimp simp add: apply_fun_pointwise_def Option.map_def split add: option.split)

  (* ---------------------------------------------------------------------- *)


text {* The operation @{text "delay v s"} delays the stream @{text s}
by one clock tick and prefixes it with the value @{text v}. The
definition is based on the definition of a function @{text
previous_defined} which fetches the latest non-$\bot$ element from a
stream, starting from a given position (and returning a value @{text
v} if there is no previous non-$\bot$ element). For example, @{text
"previous_defined x0 s 1"} applied to the stream $s = \bot \ \bot \ x1 \ \bot \ x2 \ \bot \ \bot$ returns @{text x0}, whereas @{text
"previous_defined x0 s 6"} returns $x2$.
*}


text {* Signal composition functions *}

definition
  previous_defined :: "['v, ('t::linorder, 'v) signal, 't] \<Rightarrow> 'v" where
  "previous_defined v s t == 
     (let A = {t' \<in> dom s. t' < t} in
      if A = {} then v else the (s (Max A)))"

definition
  const_on :: "['v, 't set] \<Rightarrow> ('t, 'v) signal"
where
  "const_on = (\<lambda> v A t. if t \<in> A then Some v else None)"

definition
  signalpair :: "[('t,'v::vals_cl) signal, ('t,'v) signal] \<Rightarrow> ('t,'v) signal" where
  "signalpair s s' == (\<lambda> t. (lift_pair (s t) (s' t)))"

definition
  delay :: "['v, ('t::linorder, 'v) signal] \<Rightarrow> ('t, 'v) signal" where
  "delay v s == \<lambda> t. case (s t) of 
        None \<Rightarrow> None 
      | Some v' \<Rightarrow> Some (previous_defined v s t)"

definition
  when :: "[('t,'v::vals_cl) signal, ('t,'v) signal] \<Rightarrow> ('t,'v) signal" where
  "when s s' == (\<lambda> t. 
            case (s' t) of 
              None \<Rightarrow> None 
            | Some v \<Rightarrow> if (true_val v) then (s t) else None)"
definition
  default :: "[('t,'v) signal, ('t,'v) signal] \<Rightarrow> ('t,'v) signal" where
  "default s s' == s' ++ s"


definition 
  defined_true :: "('t,'v::vals_cl) signal \<Rightarrow> 't set" where
  "defined_true s == {t \<in> dom s. true_val (the (s t))}"


  (* ---------------------------------------------------------------------- *)


text {* Domains of signal composition functions *}

lemma lift_pair_Some: "(lift_pair vo vo' = Some x) = 
      (\<exists> v v'. (x = (pair_val (v, v')) \<and> 
                vo = Some v \<and> vo' = Some v'))"
by (auto simp add: lift_pair_def split add: option.split)

lemma dom_const_on [simp]: "dom (const_on v A) = A"
by (simp add: const_on_def dom_def)

lemma dom_signalpair [simp]:
  "dom (signalpair s s') = (dom s \<inter> dom s')"
apply (simp add: dom_def signalpair_def)
apply (fastforce simp add: lift_pair_Some)
done

lemma dom_apply_fun_pointwise [simp]: 
  "dom (apply_fun_pointwise f s) = dom s"
by (simp add: dom_def apply_fun_pointwise_def)

lemma dom_delay [simp]: "dom (delay v s) = dom s"
by (simp add: dom_def delay_def split add: option.split)

lemma dom_when [simp]: "dom (when s s') = dom s \<inter> defined_true s' "
by (simp add: dom_def when_def defined_true_def split add: option.split) fastforce

lemma dom_default [simp]: "dom (default s s') = dom s \<union> dom s'"
by (simp add: dom_def default_def) fastforce



lemma defined_true_when:
  "defined_true (when s s') = (defined_true s) \<inter> (defined_true s')"
apply (simp add: defined_true_def)
apply (simp add: when_def)
apply auto
done


lemma defined_true_default:
  "defined_true (default s s') = 
     (defined_true s) \<union> (defined_true s' - dom s)"
apply (simp add: defined_true_def)
apply (simp add: default_def)
apply (simp add: map_add_def)
apply (auto split add: option.split)
done

  (* ---------------------------------------------------------------------- *)

definition 
  discrete_event_signal :: "('t::order,'v) signal \<Rightarrow> bool" where
  "discrete_event_signal s == \<forall> t. finite ({t'. t' \<le> t} \<inter> dom s)"
definition
  discrete_event_trace :: "('t::order,'n,'v) trace \<Rightarrow> bool" where
  "discrete_event_trace tr == \<forall> x. discrete_event_signal (trace_proj tr x)"

lemma discrete_event_signal_finite_dom [simp]:
  "discrete_event_signal s \<Longrightarrow> finite {t' \<in> dom s. t' \<le> t}"
apply (subgoal_tac "{t' \<in> dom s. t' \<le> t} = ({t'. t' \<le> t} \<inter> dom s)")
apply (simp add: discrete_event_signal_def)
apply fast
done

lemma discrete_event_signal_finite_dom_le [simp]:
  "discrete_event_signal s \<Longrightarrow> finite {t' \<in> dom s. t' < t}"
apply (rule finite_subset)
prefer 2
apply (erule_tac t=t in discrete_event_signal_finite_dom)
apply auto
done

lemma discrete_event_signal_signalpair [simp]:
  "\<lbrakk> discrete_event_signal s; discrete_event_signal s' \<rbrakk>
  \<Longrightarrow> discrete_event_signal (signalpair s s')"
by (clarsimp simp add: discrete_event_signal_def)
   (fastforce simp only: Int_assoc [THEN sym])

lemma discrete_event_signal_apply_fun_pointwise [simp]:
  "discrete_event_signal s 
  \<Longrightarrow> discrete_event_signal (apply_fun_pointwise f s)"
by (clarsimp simp add: discrete_event_signal_def)

lemma discrete_event_signal_delay [simp]:
  "discrete_event_signal s 
  \<Longrightarrow> discrete_event_signal (delay v s)"
by (clarsimp simp add: discrete_event_signal_def)


lemma discrete_event_signal_when [simp]:
  "\<lbrakk> discrete_event_signal s; discrete_event_signal s' \<rbrakk>
  \<Longrightarrow> discrete_event_signal (when s s')"
by (clarsimp simp add: discrete_event_signal_def)
   (fastforce simp only: Int_assoc [THEN sym])

lemma discrete_event_signal_default [simp]:
  "\<lbrakk> discrete_event_signal s; discrete_event_signal s' \<rbrakk>
  \<Longrightarrow> discrete_event_signal (default s s')"
by (clarsimp simp add: discrete_event_signal_def)
   (fastforce simp only: Int_Un_distrib)

end

