theory Streams imports Aux

begin


text {*

Our semantics is based on streams, which are essentially maps from a
temporal domain to values.  In order to combine streams, we define
first some functions on option values.  *}



(* ---------------------------------------------------------------------- *)
subsection {* Events and Traces *}

text {*

An event is a mapping from (variable) names @{text "'n"} to values
@{text "'v"}. Differently said, an event is the common behaviour of a
number of signals at a given moment of time. Here, @{text None}
indicates the absence of a value, @{text "Some v"} the presence of a
value @{text "v"}.

A @{text trace} describes the occurrence of events over time, thus
the mapping from a temporal domain @{text 't} to events.  

Altogether, the type @{text trace} is thus @{text "'t \<Rightarrow> 'n \<Rightarrow> 'v
option"}. In early experiments with the formalization, this has turned
out to be more convenient than the equivalent definition @{text "'n \<Rightarrow>
't \<Rightarrow> 'v option"}, because it permits to operate more easily on events
occurring at the same time.

A @{text signal} is the projection of a trace to one specific variable. 
*}

(* A signal is a trace restricted to one specific variable *)
type_synonym ('n, 'v) event = "'n \<Rightarrow> 'v option"
type_synonym ('t,'v) signal = "'t  \<Rightarrow>  'v option"
type_synonym ('t,'n,'v) trace = "'t  \<Rightarrow>  ('n, 'v) event"

definition
  -- "restriction of a trace on a set of variables"
  trace_restr :: "[('t,'n, 'v) trace, 'n set] \<Rightarrow> ('t,'n, 'v) trace" where
  "trace_restr tr A == \<lambda> t. (tr t)|`A"
definition
  trace_proj :: "[('t,'n, 'v) trace, 'n] \<Rightarrow> ('t, 'v) signal" where
  "trace_proj tr x == \<lambda> t. (tr t x)"
definition trace_upd :: "[('t,'n, 'v) trace, 'n, ('t, 'v) signal] \<Rightarrow> ('t,'n, 'v) trace" where
  "trace_upd tr x s == \<lambda> t. (tr t)(x:= s t)"

lemma trace_proj_trace_upd [simp]:
  "(trace_proj (trace_upd tr x s) y) = 
  (if x = y then s else trace_proj tr y)"
by (simp add: trace_proj_def trace_upd_def)

lemma trace_upd_trace_proj_same [simp]: 
  "(trace_upd tr a (trace_proj tr a)) = tr"
by (simp add: trace_upd_def trace_proj_def)


lemma trace_proj_fun_upd_map:
 "(trace_proj (\<lambda>t. fun_upd_map (tr t) ([xs [\<mapsto>] ys t, x \<mapsto> f t])) z)
  = (if x = z 
     then f 
     else (trace_proj (\<lambda> t. fun_upd_map (tr t) [xs [\<mapsto>] ys t]) z))"
by (simp add: trace_proj_def fun_upd_map_def)

lemma funs_eq_except_trace_upd [simp, rule_format]:
  "\<lbrakk> nn \<in> A; \<forall>t. funs_eq_except (s t) (s' t) A \<rbrakk> \<Longrightarrow>
 \<forall>t. funs_eq_except (s t) (trace_upd s' nn str t) A"
by (clarsimp simp add: funs_eq_except_def trace_upd_def)

lemma trace_proj_funs_eq_except:
  "\<lbrakk> \<forall>t. funs_eq_except (s t) (s' t) A; x \<notin> A \<rbrakk> \<Longrightarrow>
  trace_proj s x = trace_proj s' x"
by (simp add: funs_eq_except_def trace_proj_def)

lemma trace_proj_funs_eq_on:
  "\<lbrakk> \<forall>t. funs_eq_on (s t) (s' t) A; x \<in> A \<rbrakk> \<Longrightarrow>
  trace_proj s x = trace_proj s' x"
by (simp add: funs_eq_on_def trace_proj_def)

lemma trace_upd_map:
  "x \<notin> set ys \<Longrightarrow> map (trace_upd tr x s a) ys = map (tr a) ys"
by (simp add: trace_upd_def)


(* ---------------------------------------------------------------------- *)
subsection {* Synchronization *}


definition
  sync_signals :: "('t,'v) signal set \<Rightarrow> bool" where
  "sync_signals A == \<forall> s \<in> A. \<forall> s' \<in> A. dom s = dom s' "
definition
  sync_vars :: "[('t,'n,'v) trace, 'n set] \<Rightarrow> bool" where
  "sync_vars tr A == sync_signals ((trace_proj tr) ` A)"

lemma sync_signals_unionD: 
  "sync_signals (A \<union> B) \<Longrightarrow> (sync_signals A \<and> sync_signals B)"
by (fastforce simp add: sync_signals_def)

lemma sync_signals_insertD: 
  "sync_signals (insert x A) \<Longrightarrow> (\<forall> s' \<in> A. dom x = dom s') \<and> sync_signals A"
by (simp only: sync_signals_def)
   (blast dest: bspec)

lemma sync_signals_subset:
  "\<lbrakk> sync_signals A; B \<subseteq> A \<rbrakk> \<Longrightarrow> sync_signals B"
by (clarsimp simp add: sync_signals_def)
   (blast dest: bspec)


end


